# Copyright (C) 2019-2020 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime

import attr
from hypothesis import given, settings
import iso8601

from swh.model.hashutil import DEFAULT_ALGORITHMS
from swh.model.hypothesis_strategies import (
    aware_datetimes,
    contents,
    object_dicts,
    objects,
    origin_visits,
    persons,
    skipped_contents,
    snapshots,
)
from swh.model.model import TargetType

target_types = ("content", "directory", "revision", "release", "snapshot", "alias")
all_but_skipped_content = (
    "origin",
    "origin_visit",
    "origin_visit_status",
    "snapshot",
    "release",
    "revision",
    "directory",
    "content",
    "raw_extrinsic_metadata",
)


@given(objects(blacklist_types=()))
def test_generation(obj_type_and_obj):
    (obj_type, object_) = obj_type_and_obj
    attr.validate(object_)


@given(objects(split_content=False))
def test_generation_merged_content(obj_type_and_obj):
    # we should never generate a "skipped_content" here
    assert obj_type_and_obj[0] != "skipped_content"


@given(objects(split_content=True, blacklist_types=all_but_skipped_content))
def test_generation_split_content(obj_type_and_obj):
    # we should only generate "skipped_content"
    assert obj_type_and_obj[0] == "skipped_content"


@given(objects(blacklist_types=("origin_visit", "directory")))
def test_generation_blacklist(obj_type_and_obj):
    assert obj_type_and_obj[0] not in ("origin_visit", "directory")


def assert_nested_dict(obj):
    """Tests the object is a nested dict and contains no more class
    from swh.model.model."""
    if isinstance(obj, dict):
        for (key, value) in obj.items():
            assert isinstance(key, (str, bytes)), key
            assert_nested_dict(value)
    elif isinstance(obj, tuple):
        for value in obj:
            assert_nested_dict(value)
    elif isinstance(obj, (int, float, str, bytes, bool, type(None), datetime.datetime)):
        pass
    else:
        assert False, obj


@given(object_dicts(blacklist_types=()))
def test_dicts_generation(obj_type_and_obj):
    (obj_type, object_) = obj_type_and_obj
    assert_nested_dict(object_)
    if obj_type == "content":
        COMMON_KEYS = set(DEFAULT_ALGORITHMS) | {"length", "status", "ctime"}
        if object_["status"] == "visible":
            assert set(object_) <= COMMON_KEYS | {"data"}
        elif object_["status"] == "absent":
            assert set(object_) == COMMON_KEYS | {"reason"}
        elif object_["status"] == "hidden":
            assert set(object_) <= COMMON_KEYS | {"data"}
        else:
            assert False, object_
    elif obj_type == "release":
        assert object_["target_type"] in target_types
    elif obj_type == "snapshot":
        for branch in object_["branches"].values():
            assert branch is None or branch["target_type"] in target_types


@given(aware_datetimes())
def test_datetimes(dt):
    # Checks this doesn't raise an error, eg. about seconds in the TZ offset
    iso8601.parse_date(dt.isoformat())

    assert dt.tzinfo is not None


@given(object_dicts(split_content=False))
def test_dicts_generation_merged_content(obj_type_and_obj):
    # we should never generate a "skipped_content" here
    assert obj_type_and_obj[0] != "skipped_content"


@given(object_dicts(split_content=True, blacklist_types=all_but_skipped_content))
def test_dicts_generation_split_content(obj_type_and_obj):
    # we should only generate "skipped_content"
    assert obj_type_and_obj[0] == "skipped_content"


@given(object_dicts(blacklist_types=("release", "content")))
def test_dicts_generation_blacklist(obj_type_and_obj):
    assert obj_type_and_obj[0] not in ("release", "content")


@given(objects())
def test_model_to_dicts(obj_type_and_obj):
    _, object_ = obj_type_and_obj
    object_type = object_.object_type
    obj_dict = object_.to_dict()
    assert_nested_dict(obj_dict)
    if object_type in ("content", "skipped_content"):
        COMMON_KEYS = set(DEFAULT_ALGORITHMS) | {"length", "status"}
        if object_.ctime is not None:
            COMMON_KEYS |= {"ctime"}

        if obj_dict["status"] == "visible":
            assert set(obj_dict) == COMMON_KEYS | {"data"}
        elif obj_dict["status"] == "absent":
            assert set(obj_dict) == COMMON_KEYS | {"reason"}
        elif obj_dict["status"] == "hidden":
            assert set(obj_dict) == COMMON_KEYS | {"data"}
        else:
            assert False, obj_dict
    elif object_type == "release":
        assert obj_dict["target_type"] in target_types
    elif object_type == "snapshot":
        for branch in obj_dict["branches"].values():
            assert branch is None or branch["target_type"] in target_types


@given(contents())
def test_content_aware_datetime(cont):
    assert cont.ctime is None or cont.ctime.tzinfo is not None


@given(skipped_contents())
def test_skipped_content_aware_datetime(cont):
    assert cont.ctime is None or cont.ctime.tzinfo is not None


_min_snp_size = 10
_max_snp_size = 100


@given(snapshots(min_size=_min_snp_size, max_size=_max_snp_size))
@settings(max_examples=1)
def test_snapshots_strategy(snapshot):

    branches = snapshot.branches

    assert len(branches) >= _min_snp_size
    assert len(branches) <= _max_snp_size

    aliases = []

    # check snapshot integrity
    for name, branch in branches.items():
        assert branch is None or branch.target_type.value in target_types
        if branch is not None and branch.target_type == TargetType.ALIAS:
            aliases.append(name)
            assert branch.target in branches

    # check no cycles between aliases
    for alias in aliases:
        processed_alias = set()
        current_alias = alias
        while (
            branches[current_alias] is not None
            and branches[current_alias].target_type == TargetType.ALIAS
        ):
            assert branches[current_alias].target not in processed_alias
            processed_alias.add(current_alias)
            current_alias = branches[current_alias].target


@given(snapshots(min_size=_min_snp_size, max_size=_min_snp_size))
@settings(max_examples=1)
def test_snapshots_strategy_fixed_size(snapshot):
    assert len(snapshot.branches) == _min_snp_size


@given(origin_visits())
def test_origin_visit_aware_datetime(visit):
    assert visit.date.tzinfo is not None


@given(persons())
def test_person_do_not_look_like_anonimized(person):
    assert not (
        len(person.fullname) == 32 and person.name is None and person.email is None
    )



# Copyright (C) 2019-2021 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
import string
from typing import Sequence

from hypothesis import assume
from hypothesis.extra.dateutil import timezones
from hypothesis.strategies import (
    binary,
    booleans,
    builds,
    characters,
    composite,
    datetimes,
    dictionaries,
    from_regex,
    integers,
    just,
    lists,
    none,
    one_of,
    sampled_from,
    sets,
    text,
    tuples,
)

from .from_disk import DentryPerms
from .model import (
    BaseContent,
    Content,
    Directory,
    DirectoryEntry,
    MetadataAuthority,
    MetadataFetcher,
    ObjectType,
    Origin,
    OriginVisit,
    OriginVisitStatus,
    Person,
    RawExtrinsicMetadata,
    Release,
    Revision,
    RevisionType,
    SkippedContent,
    Snapshot,
    SnapshotBranch,
    TargetType,
    Timestamp,
    TimestampWithTimezone,
)
from .swhids import ExtendedObjectType, ExtendedSWHID

pgsql_alphabet = characters(
    blacklist_categories=("Cs",), blacklist_characters=["\u0000"]
)  # postgresql does not like these


def optional(strategy):
    return one_of(none(), strategy)


def pgsql_text():
    return text(alphabet=pgsql_alphabet)


def sha1_git():
    return binary(min_size=20, max_size=20)


def sha1():
    return binary(min_size=20, max_size=20)


def binaries_without_bytes(blacklist: Sequence[int]):
    """Like hypothesis.strategies.binary, but takes a sequence of bytes that
    should not be included."""
    return lists(sampled_from([i for i in range(256) if i not in blacklist])).map(bytes)


@composite
def extended_swhids(draw):
    object_type = draw(sampled_from(ExtendedObjectType))
    object_id = draw(sha1_git())
    return ExtendedSWHID(object_type=object_type, object_id=object_id)


def aware_datetimes():
    # datetimes in Software Heritage are not used for software artifacts
    # (which may be much older than 2000), but only for objects like scheduler
    # task runs, and origin visits, which were created by Software Heritage,
    # so at least in 2015.
    # We're forbidding old datetimes, because until 1956, many timezones had seconds
    # in their "UTC offsets" (see
    # <https://en.wikipedia.org/wiki/Time_zone#Worldwide_time_zones>), which is not
    # encodable in ISO8601; and we need our datetimes to be ISO8601-encodable in the
    # RPC protocol
    min_value = datetime.datetime(2000, 1, 1, 0, 0, 0)
    return datetimes(min_value=min_value, timezones=timezones())


@composite
def iris(draw):
    protocol = draw(sampled_from(["git", "http", "https", "deb"]))
    domain = draw(from_regex(r"\A([a-z]([a-z0-9é🏛️-]*)\.){1,3}([a-z0-9é])+\Z"))

    return "%s://%s" % (protocol, domain)


@composite
def persons_d(draw):
    fullname = draw(binary())
    email = draw(optional(binary()))
    name = draw(optional(binary()))
    assume(not (len(fullname) == 32 and email is None and name is None))
    return dict(fullname=fullname, name=name, email=email)


def persons():
    return persons_d().map(Person.from_dict)


def timestamps_d():
    max_seconds = datetime.datetime.max.replace(
        tzinfo=datetime.timezone.utc
    ).timestamp()
    min_seconds = datetime.datetime.min.replace(
        tzinfo=datetime.timezone.utc
    ).timestamp()
    return builds(
        dict,
        seconds=integers(min_seconds, max_seconds),
        microseconds=integers(0, 1000000),
    )


def timestamps():
    return timestamps_d().map(Timestamp.from_dict)


@composite
def timestamps_with_timezone_d(
    draw,
    timestamp=timestamps_d(),
    offset=integers(min_value=-14 * 60, max_value=14 * 60),
    negative_utc=booleans(),
):
    timestamp = draw(timestamp)
    offset = draw(offset)
    negative_utc = draw(negative_utc)
    assume(not (negative_utc and offset))
    return dict(timestamp=timestamp, offset=offset, negative_utc=negative_utc)


timestamps_with_timezone = timestamps_with_timezone_d().map(
    TimestampWithTimezone.from_dict
)


def origins_d():
    return builds(dict, url=iris())


def origins():
    return origins_d().map(Origin.from_dict)


def origin_visits_d():
    return builds(
        dict,
        visit=integers(1, 1000),
        origin=iris(),
        date=aware_datetimes(),
        type=pgsql_text(),
    )


def origin_visits():
    return origin_visits_d().map(OriginVisit.from_dict)


def metadata_dicts():
    return dictionaries(pgsql_text(), pgsql_text())


def origin_visit_statuses_d():
    return builds(
        dict,
        visit=integers(1, 1000),
        origin=iris(),
        type=optional(sampled_from(["git", "svn", "pypi", "debian"])),
        status=sampled_from(
            ["created", "ongoing", "full", "partial", "not_found", "failed"]
        ),
        date=aware_datetimes(),
        snapshot=optional(sha1_git()),
        metadata=optional(metadata_dicts()),
    )


def origin_visit_statuses():
    return origin_visit_statuses_d().map(OriginVisitStatus.from_dict)


@composite
def releases_d(draw):
    target_type = sampled_from([x.value for x in ObjectType])
    name = binary()
    message = optional(binary())
    synthetic = booleans()
    target = sha1_git()
    metadata = optional(revision_metadata())

    d = draw(
        one_of(
            # None author/date:
            builds(
                dict,
                name=name,
                message=message,
                synthetic=synthetic,
                author=none(),
                date=none(),
                target=target,
                target_type=target_type,
                metadata=metadata,
            ),
            # non-None author/date:
            builds(
                dict,
                name=name,
                message=message,
                synthetic=synthetic,
                date=timestamps_with_timezone_d(),
                author=persons_d(),
                target=target,
                target_type=target_type,
                metadata=metadata,
            ),
            # it is also possible for date to be None but not author, but let's not
            # overwhelm hypothesis with this edge case
        )
    )

    raw_manifest = draw(optional(binary()))
    if raw_manifest:
        d["raw_manifest"] = raw_manifest
    return d


def releases():
    return releases_d().map(Release.from_dict)


revision_metadata = metadata_dicts


def extra_headers():
    return lists(
        tuples(binary(min_size=0, max_size=50), binary(min_size=0, max_size=500))
    ).map(tuple)


@composite
def revisions_d(draw):
    d = draw(
        one_of(
            # None author/committer/date/committer_date
            builds(
                dict,
                message=optional(binary()),
                synthetic=booleans(),
                author=none(),
                committer=none(),
                date=none(),
                committer_date=none(),
                parents=tuples(sha1_git()),
                directory=sha1_git(),
                type=sampled_from([x.value for x in RevisionType]),
                metadata=optional(revision_metadata()),
                extra_headers=extra_headers(),
            ),
            # non-None author/committer/date/committer_date
            builds(
                dict,
                message=optional(binary()),
                synthetic=booleans(),
                author=persons_d(),
                committer=persons_d(),
                date=timestamps_with_timezone_d(),
                committer_date=timestamps_with_timezone_d(),
                parents=tuples(sha1_git()),
                directory=sha1_git(),
                type=sampled_from([x.value for x in RevisionType]),
                metadata=optional(revision_metadata()),
                extra_headers=extra_headers(),
            ),
            # There are many other combinations, but let's not overwhelm hypothesis
            # with these edge cases
        )
    )
    # TODO: metadata['extra_headers'] can have binary keys and values

    raw_manifest = draw(optional(binary()))
    if raw_manifest:
        d["raw_manifest"] = raw_manifest
    return d


def revisions():
    return revisions_d().map(Revision.from_dict)


def directory_entries_d():
    return one_of(
        builds(
            dict,
            name=binaries_without_bytes(b"/"),
            target=sha1_git(),
            type=just("file"),
            perms=one_of(
                integers(min_value=0o100000, max_value=0o100777),  # regular file
                integers(min_value=0o120000, max_value=0o120777),  # symlink
            ),
        ),
        builds(
            dict,
            name=binaries_without_bytes(b"/"),
            target=sha1_git(),
            type=just("dir"),
            perms=integers(
                min_value=DentryPerms.directory,
                max_value=DentryPerms.directory + 0o777,
            ),
        ),
        builds(
            dict,
            name=binaries_without_bytes(b"/"),
            target=sha1_git(),
            type=just("rev"),
            perms=integers(
                min_value=DentryPerms.revision,
                max_value=DentryPerms.revision + 0o777,
            ),
        ),
    )


def directory_entries():
    return directory_entries_d().map(DirectoryEntry)


@composite
def directories_d(draw):
    d = draw(builds(dict, entries=tuples(directory_entries_d())))

    raw_manifest = draw(optional(binary()))
    if raw_manifest:
        d["raw_manifest"] = raw_manifest
    return d


def directories():
    return directories_d().map(Directory.from_dict)


def contents_d():
    return one_of(present_contents_d(), skipped_contents_d())


def contents():
    return one_of(present_contents(), skipped_contents())


def present_contents_d():
    return builds(
        dict,
        data=binary(max_size=4096),
        ctime=optional(aware_datetimes()),
        status=one_of(just("visible"), just("hidden")),
    )


def present_contents():
    return present_contents_d().map(lambda d: Content.from_data(**d))


@composite
def skipped_contents_d(draw):
    result = BaseContent._hash_data(draw(binary(max_size=4096)))
    result.pop("data")
    nullify_attrs = draw(
        sets(sampled_from(["sha1", "sha1_git", "sha256", "blake2s256"]))
    )
    for k in nullify_attrs:
        result[k] = None
    result["reason"] = draw(pgsql_text())
    result["status"] = "absent"
    result["ctime"] = draw(optional(aware_datetimes()))
    return result


def skipped_contents():
    return skipped_contents_d().map(SkippedContent.from_dict)


def branch_names():
    return binary(min_size=1)


def branch_targets_object_d():
    return builds(
        dict,
        target=sha1_git(),
        target_type=sampled_from(
            [x.value for x in TargetType if x.value not in ("alias",)]
        ),
    )


def branch_targets_alias_d():
    return builds(
        dict, target=sha1_git(), target_type=just("alias")
    )  # TargetType.ALIAS.value))


def branch_targets_d(*, only_objects=False):
    if only_objects:
        return branch_targets_object_d()
    else:
        return one_of(branch_targets_alias_d(), branch_targets_object_d())


def branch_targets(*, only_objects=False):
    return builds(SnapshotBranch.from_dict, branch_targets_d(only_objects=only_objects))


@composite
def snapshots_d(draw, *, min_size=0, max_size=100, only_objects=False):
    branches = draw(
        dictionaries(
            keys=branch_names(),
            values=optional(branch_targets_d(only_objects=only_objects)),
            min_size=min_size,
            max_size=max_size,
        )
    )

    if not only_objects:
        # Make sure aliases point to actual branches
        unresolved_aliases = {
            branch: target["target"]
            for branch, target in branches.items()
            if (
                target
                and target["target_type"] == "alias"
                and target["target"] not in branches
            )
        }
        for alias_name, alias_target in unresolved_aliases.items():
            # Override alias branch with one pointing to a real object
            # if max_size constraint is reached
            alias = alias_target if len(branches) < max_size else alias_name
            branches[alias] = draw(branch_targets_d(only_objects=True))

    # Ensure no cycles between aliases
    while True:
        try:
            snapshot = Snapshot.from_dict(
                {
                    "branches": {
                        name: branch or None for (name, branch) in branches.items()
                    }
                }
            )
        except ValueError as e:
            for (source, target) in e.args[1]:
                branches[source] = draw(branch_targets_d(only_objects=True))
        else:
            break

    return snapshot.to_dict()


def snapshots(*, min_size=0, max_size=100, only_objects=False):
    return snapshots_d(
        min_size=min_size, max_size=max_size, only_objects=only_objects
    ).map(Snapshot.from_dict)


def metadata_authorities():
    return builds(MetadataAuthority, url=iris(), metadata=just(None))


def metadata_fetchers():
    return builds(
        MetadataFetcher,
        name=text(min_size=1, alphabet=string.printable),
        version=text(
            min_size=1,
            alphabet=string.ascii_letters + string.digits + string.punctuation,
        ),
        metadata=just(None),
    )


def raw_extrinsic_metadata():
    return builds(
        RawExtrinsicMetadata,
        target=extended_swhids(),
        discovery_date=aware_datetimes(),
        authority=metadata_authorities(),
        fetcher=metadata_fetchers(),
        format=text(min_size=1, alphabet=string.printable),
    )


def raw_extrinsic_metadata_d():
    return raw_extrinsic_metadata().map(RawExtrinsicMetadata.to_dict)


def objects(blacklist_types=("origin_visit_status",), split_content=False):
    """generates a random couple (type, obj)

    which obj is an instance of the Model class corresponding to obj_type.

    `blacklist_types` is a list of obj_type to exclude from the strategy.

    If `split_content` is True, generates Content and SkippedContent under different
    obj_type, resp. "content" and "skipped_content".
    """
    strategies = [
        ("origin", origins),
        ("origin_visit", origin_visits),
        ("origin_visit_status", origin_visit_statuses),
        ("snapshot", snapshots),
        ("release", releases),
        ("revision", revisions),
        ("directory", directories),
        ("raw_extrinsic_metadata", raw_extrinsic_metadata),
    ]
    if split_content:
        strategies.append(("content", present_contents))
        strategies.append(("skipped_content", skipped_contents))
    else:
        strategies.append(("content", contents))
    args = [
        obj_gen().map(lambda x, obj_type=obj_type: (obj_type, x))
        for (obj_type, obj_gen) in strategies
        if obj_type not in blacklist_types
    ]
    return one_of(*args)


def object_dicts(blacklist_types=("origin_visit_status",), split_content=False):
    """generates a random couple (type, dict)

    which dict is suitable for <ModelForType>.from_dict() factory methods.

    `blacklist_types` is a list of obj_type to exclude from the strategy.

    If `split_content` is True, generates Content and SkippedContent under different
    obj_type, resp. "content" and "skipped_content".

    """
    strategies = [
        ("origin", origins_d),
        ("origin_visit", origin_visits_d),
        ("origin_visit_status", origin_visit_statuses_d),
        ("snapshot", snapshots_d),
        ("release", releases_d),
        ("revision", revisions_d),
        ("directory", directories_d),
        ("raw_extrinsic_metadata", raw_extrinsic_metadata_d),
    ]
    if split_content:
        strategies.append(("content", present_contents_d))
        strategies.append(("skipped_content", skipped_contents_d))
    else:
        strategies.append(("content", contents_d))
    args = [
        obj_gen().map(lambda x, obj_type=obj_type: (obj_type, x))
        for (obj_type, obj_gen) in strategies
        if obj_type not in blacklist_types
    ]
    return one_of(*args)



# Copyright (C) 2017 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from operator import itemgetter
import os
import sys

from swh.model.from_disk import DentryPerms, Directory
from swh.model.hashutil import ALGORITHMS, hash_to_hex


def generate_from_directory(varname, directory, indent=0):
    """Generate test data from a given directory"""

    def get_data(member, path):
        yield (path, member.get_data())
        if isinstance(member, Directory):
            for name, child in member.items():
                yield from get_data(child, os.path.join(path, name))

    data = dict(get_data(directory, b""))
    out = []

    def format_hash(h, indent=0):
        spindent = " " * indent
        if len(h) > 20:
            cutoff = len(h) // 2
            parts = h[:cutoff], h[cutoff:]
        else:
            parts = [h]

        out.append("hash_to_bytes(\n")
        for part in parts:
            out.append(spindent + "    %s\n" % repr(hash_to_hex(part)))
        out.append(spindent + ")")

    def format_dict_items(d, indent=0):
        spindent = " " * indent
        for key, value in sorted(d.items()):
            if isinstance(key, bytes):
                out.append(spindent + repr(key) + ": {\n")
                format_dict_items(value, indent=indent + 4)
                out.append(spindent + "}")
            else:
                out.append(spindent + repr(key) + ": ")
                if key == "entries":
                    if not value:
                        out.append("[]")
                    else:
                        out.append("[")
                        last_index = len(value) - 1
                        for i, entry in enumerate(
                            sorted(value, key=itemgetter("name"))
                        ):
                            if i:
                                out.append(" ")
                            out.append("{\n")
                            format_dict_items(entry, indent=indent + 4)
                            if i != last_index:
                                out.append(spindent + "},")
                        out.append(spindent + "}]")
                elif key in ALGORITHMS | {"id", "target"}:
                    format_hash(value, indent=indent)
                elif isinstance(value, DentryPerms):
                    out.append(str(value))
                else:
                    out.append(repr(value))
            out.append(",\n")

    spindent = " " * indent
    out.append(spindent + "%s = {\n" % varname)
    format_dict_items(data, indent=4 + indent)
    out.append(spindent + "}")

    return "".join(out)


if __name__ == "__main__":
    if not sys.argv[1:]:
        print("Usage: %s dir1 dir2" % sys.argv[0], file=sys.stderr)
        exit(2)

    for dirname in sys.argv[1:]:
        basename = os.path.basename(dirname)
        varname = "expected_%s" % basename
        testdata = generate_from_directory(
            varname, Directory.from_disk(path=os.fsencode(dirname)), indent=8
        )
        print(testdata)
        print()



# Copyright (C) 2015  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
import unittest

from swh.model.exceptions import ValidationError
from swh.model.fields import simple


class ValidateSimple(unittest.TestCase):
    def setUp(self):
        self.valid_str = "I am a valid string"

        self.valid_bytes = b"I am a valid bytes object"

        self.enum_values = {"an enum value", "other", "and another"}
        self.invalid_enum_value = "invalid enum value"

        self.valid_int = 42

        self.valid_real = 42.42

        self.valid_datetime = datetime.datetime(
            1999, 1, 1, 12, 0, 0, tzinfo=datetime.timezone.utc
        )
        self.invalid_datetime_notz = datetime.datetime(1999, 1, 1, 12, 0, 0)

    def test_validate_int(self):
        self.assertTrue(simple.validate_int(self.valid_int))

    def test_validate_int_invalid_type(self):
        with self.assertRaises(ValidationError) as cm:
            simple.validate_int(self.valid_str)

        exc = cm.exception
        self.assertIsInstance(str(exc), str)
        self.assertEqual(exc.code, "unexpected-type")
        self.assertEqual(exc.params["expected_type"], "Integral")
        self.assertEqual(exc.params["type"], "str")

    def test_validate_str(self):
        self.assertTrue(simple.validate_str(self.valid_str))

    def test_validate_str_invalid_type(self):
        with self.assertRaises(ValidationError) as cm:
            simple.validate_str(self.valid_int)

        exc = cm.exception
        self.assertIsInstance(str(exc), str)
        self.assertEqual(exc.code, "unexpected-type")
        self.assertEqual(exc.params["expected_type"], "str")
        self.assertEqual(exc.params["type"], "int")

        with self.assertRaises(ValidationError) as cm:
            simple.validate_str(self.valid_bytes)

        exc = cm.exception
        self.assertIsInstance(str(exc), str)
        self.assertEqual(exc.code, "unexpected-type")
        self.assertEqual(exc.params["expected_type"], "str")
        self.assertEqual(exc.params["type"], "bytes")

    def test_validate_bytes(self):
        self.assertTrue(simple.validate_bytes(self.valid_bytes))

    def test_validate_bytes_invalid_type(self):
        with self.assertRaises(ValidationError) as cm:
            simple.validate_bytes(self.valid_int)

        exc = cm.exception
        self.assertIsInstance(str(exc), str)
        self.assertEqual(exc.code, "unexpected-type")
        self.assertEqual(exc.params["expected_type"], "bytes")
        self.assertEqual(exc.params["type"], "int")

        with self.assertRaises(ValidationError) as cm:
            simple.validate_bytes(self.valid_str)

        exc = cm.exception
        self.assertIsInstance(str(exc), str)
        self.assertEqual(exc.code, "unexpected-type")
        self.assertEqual(exc.params["expected_type"], "bytes")
        self.assertEqual(exc.params["type"], "str")

    def test_validate_datetime(self):
        self.assertTrue(simple.validate_datetime(self.valid_datetime))
        self.assertTrue(simple.validate_datetime(self.valid_int))
        self.assertTrue(simple.validate_datetime(self.valid_real))

    def test_validate_datetime_invalid_type(self):
        with self.assertRaises(ValidationError) as cm:
            simple.validate_datetime(self.valid_str)

        exc = cm.exception
        self.assertIsInstance(str(exc), str)
        self.assertEqual(exc.code, "unexpected-type")
        self.assertEqual(exc.params["expected_type"], "one of datetime, Real")
        self.assertEqual(exc.params["type"], "str")

    def test_validate_datetime_invalide_tz(self):
        with self.assertRaises(ValidationError) as cm:
            simple.validate_datetime(self.invalid_datetime_notz)

        exc = cm.exception
        self.assertIsInstance(str(exc), str)
        self.assertEqual(exc.code, "datetime-without-tzinfo")

    def test_validate_enum(self):
        for value in self.enum_values:
            self.assertTrue(simple.validate_enum(value, self.enum_values))

    def test_validate_enum_invalid_value(self):
        with self.assertRaises(ValidationError) as cm:
            simple.validate_enum(self.invalid_enum_value, self.enum_values)

        exc = cm.exception
        self.assertIsInstance(str(exc), str)
        self.assertEqual(exc.code, "unexpected-value")
        self.assertEqual(exc.params["value"], self.invalid_enum_value)
        self.assertEqual(
            exc.params["expected_values"], ", ".join(sorted(self.enum_values))
        )



# Copyright (C) 2017-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import unittest

from swh.model.toposort import toposort


def is_toposorted_slow(revision_log):
    """Check (inefficiently) that the given revision log is in any topological
    order.

    Complexity: O(n^2).
        (Note: It's totally possible to write a O(n) is_toposorted function,
        but it requires computing the transitive closure of the input DAG,
        which requires computing a topological ordering of that DAG, which
        kind of defeats the purpose of writing unit tests for toposort().)

    Args:
        revision_log: Revision log as returned by
            swh.storage.Storage.revision_log().

    Returns:
        True if the revision log is topologically sorted.
    """
    rev_by_id = {r["id"]: r for r in revision_log}

    def all_parents(revision):
        for parent in revision["parents"]:
            yield parent
            yield from all_parents(rev_by_id[parent])

    visited = set()
    for rev in revision_log:
        visited.add(rev["id"])
        if not all(parent in visited for parent in all_parents(rev)):
            return False
    return True


class TestToposort(unittest.TestCase):
    def generate_log(self, graph):
        for node_id, parents in graph.items():
            yield {"id": node_id, "parents": tuple(parents)}

    def unordered_log(self, log):
        return {(d["id"], tuple(d["parents"])) for d in log}

    def check(self, graph):
        log = list(self.generate_log(graph))
        topolog = list(toposort(log))
        self.assertEqual(len(topolog), len(graph))
        self.assertEqual(self.unordered_log(topolog), self.unordered_log(log))
        self.assertTrue(is_toposorted_slow(toposort(log)))

    def test_linked_list(self):
        self.check({3: [2], 2: [1], 1: []})

    def test_fork(self):
        self.check({7: [6], 6: [4], 5: [3], 4: [2], 3: [2], 2: [1], 1: []})

    def test_fork_merge(self):
        self.check({8: [7, 5], 7: [6], 6: [4], 5: [3], 4: [2], 3: [2], 2: [1], 1: []})

    def test_two_origins(self):
        self.check({9: [8], 8: [7, 5], 7: [6], 6: [4], 5: [3], 4: [], 3: []})

    def test_three_way(self):
        self.check(
            {
                9: [8, 4, 2],
                8: [7, 5],
                7: [6],
                6: [4],
                5: [3],
                4: [2],
                3: [2],
                2: [1],
                1: [],
            }
        )
